import { StagehandContainer } from "./StagehandContainer";
/**
 * Decide which container to create.
 */
export declare function createStagehandContainer(obj: Window | HTMLElement): StagehandContainer;
