export declare class PlaywrightCommandException extends Error {
    constructor(message: string);
}
export declare class PlaywrightCommandMethodNotSupportedException extends Error {
    constructor(message: string);
}
export interface GotoOptions {
    timeout?: number;
    waitUntil?: "load" | "domcontentloaded" | "networkidle" | "commit";
    referer?: string;
}
