export type TextAnnotation = {
    text: string;
    bottom_left: {
        x: number;
        y: number;
    };
    bottom_left_normalized: {
        x: number;
        y: number;
    };
    width: number;
    height: number;
};
