export declare function waitForDomSettle(): Promise<void>;
export declare function calculateViewportHeight(): number;
/**
 * Tests if the element actually responds to .scrollTo(...)
 * and that scrollTop changes as expected.
 */
export declare function canElementScroll(elem: HTMLElement): boolean;
export declare function getNodeFromXpath(xpath: string): Node;
