/**
 * This file is meant to be used as a scratchpad for trying out actionable observe.
 * To create a Stagehand project with best practices and configuration, run:
 *
 * npx create-browser-app@latest my-browser-app
 */
export {};
